package Tie::PerlFiles;

use strict;
use vars qw($VERSION @ISA);

use Carp;
use Data::Dumper;
use Exporter;

@ISA = qw(Exporter);

# set this to make more compact datafiles
$Data::Dumper::Indent=1;
# set this to fix any self-referential structures
$Data::Dumper::Purity=1;

$VERSION = '1.3';

sub whowasi { (caller(1))[3] . '(' . join(',', @_) . ')' }

my $handlect = 0;

# use to set debug message reporting
#
my $DEBUG=0;
sub debug { 
   my $self = shift;
   $DEBUG = @_ ? shift : $DEBUG;
   return $DEBUG;
}

# set the size of the datacache (number of key/value pairs to keep)
#
my $STACKSIZE=25;
sub stacksize { 
   my $self = shift;
   $STACKSIZE = (@_ ? shift : $STACKSIZE); 
   $self->_push_stack();  # will truncate CACHE if needed
   return $STACKSIZE; 
}

# use to change read vs read/write mode
#
sub clobber {
   my $self = shift;
   $self->{CLOBBER} = @_ ? shift : $self->{CLOBBER};
   return $self->{CLOBBER};
}

# use these routines to enforce a lock on this dataset
#
sub lock {
   carp &whowasi if $DEBUG;
   my $self = shift;
   croak("@{[&whowasi]}: Can't lock, no lockfile defined")
      if ($self->{LOCKFILE} eq "");
   open($self->{LOCKHANDLE},">".$self->{LOCKFILE})
      || croak("@{[&whowasi]}: Can't create lockfile: $!");
   flock($self->{LOCKHANDLE}, 2); # exclusive lock;
}

sub unlock {
   carp &whowasi if $DEBUG;
   my $self = shift;
   flock($self->{LOCKHANDLE}, 8); # unlock
   close($self->{LOCKHANDLE});    # don't need that anymore
}

# use this routine to flush the CACHE, as needed
#
sub flush {
   carp &whowasi if $DEBUG;
   my $self = shift;
   my $key = shift;  

   if ($key) { # clear this one key
      delete $self->{CACHE}->{$key};
      my $i=0;
      while ($i <= $#{$self->{STACK}}) {
         if ($key eq $self->{STACK}->[$i]) {
            splice @{$self->{STACK}}, $i, 1;
         } else {
            $i++;
         }
      }
   } else { # clear entire CACHE
      # STACK should contain all the keys, but lets be safe
      my @keys = keys %{$self->{CACHE}};
      foreach my $k (@keys) {
         delete $self->{CACHE}->{$k};
      }
      @{$self->{STACK}} = ();
   }
}

# constructor for the tied class
#
sub TIEHASH {
   carp &whowasi if $DEBUG;
   my $self = shift;
   my $path = shift;
   my $mode = shift || "r";
   my $lockfile = shift;

   # too many params
   croak("Usage: @{[&whowasi]} Path [[Mode] Lockfile]") if (@_);

   # some debugging
   if ($DEBUG > 1) {
      carp "@{[&whowasi]} Path = $path";
      carp "@{[&whowasi]} Mode = $mode";
      carp "@{[&whowasi]} Lock = $lockfile";
   }

   # clean path, don't want trailing slash
   $path =~ s#/$##;

   # set write mode
   my $clobber = ($mode eq "rw" ? 1 : 0);

   # check for existing data directory
   unless (-e $path && -d $path) {
      if ($clobber) { # we can try to create dir if needed
         unless (mkdir($path, 0755)) {
            croak("@{[&whowasi]}: Can't create $path: $!");
         }
      } else { # we cannot create dir
         croak("@{[&whowasi]}: $path does not exist");
      }
   }

   # make up two handles, for directory and lockfile
   # similar to code in Symbol.pm
   my ($lockhandle, $dirhandle);
   {
      no strict 'refs';
      $lockhandle = \*{"Tie::PerlFiles::HANDLE".$handlect++};
      $dirhandle = \*{"Tie::PerlFiles::HANDLE".$handlect++};
   }

   # save everything for this object
   my $node = {
      PATH => $path,
      HANDLE => $dirhandle,
      LOCKFILE => $lockfile,
      LOCKHANDLE => $lockhandle,
      CACHE => {},
      STACK => [],
      CLOBBER => $clobber,
   };

   # open directory for future use
   opendir($node->{HANDLE}, $path) 
      || croak("@{[&whowasi]}: Can't open $path: $!");

   return bless $node, $self;
}

# called whenever a key is looked up, returns the value
#
sub FETCH {
   carp &whowasi if $DEBUG;
   my $self = shift;
   my $key = shift;
   my $file = $self->{PATH} . "/" . &_safe_filename($key);

   return undef if !&_key_ok($key);

   # return cached value, if we have one
   if (exists $self->{CACHE}->{$key}) {
      my $mtime = (stat($file))[9];
      if ($mtime <= $self->{CACHE}->{$key}->[0]) {
         return $self->{CACHE}->{$key}->[1];  # return the value
      }
   }

   # grab code from file and eval it
   my $fh;
   {
      no strict 'refs';
      $fh = \*{"Tie::PerlFiles::HANDLE".$handlect++};
   }
   open($fh, "$file") || return undef;
   local $/ = undef; 
   my $code = scalar <$fh>;
   close($fh);
   my $data = "";
   my $res = eval($code);
   return undef unless $res;
   my $mtime = (stat($file))[9];
   $self->{CACHE}->{$key} = [$mtime, $data];
   $self->_push_stack($key);

   # return the referenced hash
   return $data;
}

# called when a value is to be stored
#
sub STORE {
   carp &whowasi if $DEBUG;
   my $self = shift;
   my $key = shift;
   my $value = shift;
   my $file = $self->{PATH} . "/" . &_safe_filename($key);
   my $tfile = $self->{PATH} . "/," . &_safe_filename($key) . ".$$";

   if (!&_key_ok($key)) {
      carp("@{[&whowasi]}: Bad key '$key'");
      return;
   }

   unless ($self->{CLOBBER}) {
      carp("@{[&whowasi]}: No write access for $self->{'PATH'}");
      return;
   }

   local *FH;
   unless (open(FH,">$tfile")) {
      croak("@{[&whowasi]}: Can't open $tfile: $!");
      return;
   }
   my $dump = Data::Dumper->Dump([$value],['data']);
   print FH $dump;
   print FH "\n1;\n";
   close(FH);

   unless (rename $tfile, $file) {
      croak("@{[&whowasi]}: Can't save to $file: $!");
      return;
   }

   # update the CACHE
   my $mtime = (stat($file))[9];
   $self->{CACHE}->{$key} = [$mtime, $value];
   $self->_push_stack($key);
}

# called when a value is to be removed
#
sub DELETE {
   carp &whowasi if $DEBUG;
   my $self = shift;
   my $key = shift;
   my $file = $self->{PATH} . "/" . &_safe_filename($key);

   if (!&_key_ok($key)) {
      carp("@{[&whowasi]}: Bad key '$key'");
      return;
   }

   croak("@{[&whowasi]}: Can't remove file $file") 
      unless $self->{CLOBBER};

   unless (unlink $file) {
      carp("@{[&whowasi]}: Can't unlink $file: $!");
      return;
   }
   delete $self->{CACHE}->{$key};
}

# called when hash is to be zero'ed
#
sub CLEAR {
   carp &whowasi if $DEBUG;
   croak("@{[&whowasi]}: No! I won't let you clear this.");
}

# called to determine if a given key is defined
#
sub EXISTS {
   carp &whowasi if $DEBUG;
   my $self = shift;
   my $key = shift;
   my $file = $self->{PATH} . "/" . &_safe_filename($key);

   return undef if (!&_key_ok($key));
   return undef if ( -d $file );
   return ( -e $file );
}

# called when tie is closed
#
sub DESTROY {
   carp &whowasi if $DEBUG;
   closedir(shift()->{HANDLE});
}

# called whenever a list of keys, values, or pairs is to be made
#
sub FIRSTKEY {
   carp &whowasi if $DEBUG;   
   my $self = shift;

   my $entry;
   rewinddir($self->{HANDLE});
   while ($entry = readdir($self->{HANDLE})) {
      next if ( -d  $self->{PATH}.'/'.$entry );
      $entry = &_decode_key($entry);
      return $entry if (&_key_ok($entry));
   }
   return;
}

# called after FIRSTKEY to get each additional key
sub NEXTKEY {
   carp &whowasi if $DEBUG;
   my $self = shift;

   my $entry;
   while (defined($entry = readdir($self->{HANDLE}))) {
      next if ( -d  $self->{PATH}.'/'.$entry );
      $entry = &_decode_key($entry);
      return $entry if (&_key_ok($entry));
   }
   return;
}

#################################################
# Internal methods

# restrict names of keys here
#
sub _key_ok {
   return 0 if ($_[0] =~ /^$/); # keys must contain at least one character
   return 1;
}

# generate a safe filename for the key
#
sub _safe_filename {
   my $key = shift;

   $key =~ s/(\W)/sprintf("%%%x", ord($1))/eg;
   return $key;
}

# restore key name from filename
#
sub _decode_key {
   my $key = shift;
   return "" if ($key !~ /^[%\w]+$/);
   $key =~ s/%([a-fA-F\d]{2})/pack("C",hex($1))/eg;
   return $key;
}

# manage the stack to limit memory usage
#
sub _push_stack {
   my $self = shift;
   my $key = shift;

   if ($key) { # skip push if no key is given

      # push current entry to newest position
      my $i=0;
      while ($i <= $#{$self->{STACK}}) {
         if ($key eq $self->{STACK}->[$i]) {
            splice @{$self->{STACK}}, $i, 1;
         } else { 
            $i++; 
         }
      }
      push @{$self->{STACK}}, $key;
   }

   # remove oldest entries if stack is full
   while ($#{$self->{STACK}} >= $STACKSIZE) {
      my $old = shift @{$self->{STACK}};
      delete $self->{CACHE}->{$old};
   }

   carp("STACK contains: ".join(',',@{$self->{STACK}})) if $DEBUG;
}

# always return a true
1;

__END__

=head1 NAME

Tie::PerlFiles - interface to directory of eval-able datafiles

=head1 SYNOPSIS

 use Tie::PerlFiles;

 # tie hash to files in /data/dir, using read/write mode, 
 #     with common lockfile /my/lockfile
 tie (%hash, 'Tie::PerlFiles', '/data/dir', 'rw', '/my/lockfile');

 # create /data/dir/scale containing a scalar variable with the
 # data 'one two three'
 $hash{'scale'} = 'one two three';

 # create a blessed, complex data item
 $ref = [ 'one', 'two', { 'a' => 10, 'b' => 20 } ];
 bless $ref, "Goober";

 # save the complex data item using locking
 tied(%hash)->lock();
 $hash{'complex'} = $ref;
 tied(%hash)->unlock();

 # finished
 untie %hash;

=head1 DESCRIPTION

This module provides the ability to C<tie> a directory of eval-able
text files to a hash variable. The module uses C<Data::Dumper> to
create the datafiles, which allows you to store almost any complex
data structure as a hash value while preserving the structure.

B<Tie::PerlFiles> includes a simple caching mechanism. By default, the
last 25 key-value pairs are stored in memory. The cache automatically
removes the oldest key-value pairs when the space is needed for new
data. The cache mechanism checks the timestamp of the stored file,
and automatically reloads the requested item if the saved copy is 
newer. Note that this is a comparison of unixtimes, which have a 
one second granularity.

B<Tie::PerlFiles> also supports a simple file locking mechanism. If
you provide the name of a lockfile, you can call a C<lock()> functions
on the tied hash before making modifications, to ensure exclusive
access.

=head1 USAGE

=over

=item tie (%hash, 'Tie::PerlFiles', data-dir [, mode [, lockfile ] ])

This ties the I<%hash> variable to the files contained in 
the directory I<data-dir>. I<mode> should be 'rw' for read/write access,
or anything else for read-only access. I<lockfile> should be the name
of a scratch file to be used for locking, if you plan to use the 
included C<lock()> and C<unlock()> functions.

=item tied(%hash)->debug( [value] )

Use this function to set the I<DEBUG> parameter to I<value>. 
Debugging messages are reported for any non-zero value.
If called with no parameters, this function returns the 
current I<DEBUG> value.

=item tied(%hash)->stacksize( [value] )

Use this function to set the I<STACKSIZE> parameter to I<value>.
This parameter is the number of key-value pairs that are stored in the
memory cache by this module. If you set this value below the current
usage of the cache, the oldest entries will be cleared from the cache.
If called with no parameters, this function returns the
current I<STACKSIZE> value.

=item tied(%hash)->clobber( [value] )

Use this function to set the I<CLOBBER> parameter to I<value>.  The
parameter sets the access mode to read-only (I<CLOBBER>=0) or
read-write (I<CLOBBER>=1). If called with no parameters, this function
returns the current I<CLOBBER> value.

=item tied(%hash)->lock()

Use this function to wait for and create an exclusive lock on the 
provided I<lockfile>. The module will open the given lockfile for writing
and then wait to establish a lock using C<flock()>. The I<lockfile>
must be provided when the has is tied, and the file must be writable
by the process, or the C<lock()> will fail.

=item tied(%hash)->unlock()

Use this function to remove the lock created by the C<lock()>
function, and to close the I<lockfile>.

=item tied(%hash)->flush( [key] )

Use this function to remove key-value pairs from the memory cache.
If a I<key> is provided, then that key-value pair will be
removed from the cache. If no I<key> is provided, then entire cache 
will be cleared.

=back

=head1 CAUTIONS

Keys can have any non-null values. All keys are HTML-encoded 
before used as filenames to avoid dangerous misinterpretation
of the key as a directory path. (i.e. the key '../../etc/passwd'
could not be used to try to access etc/passwd from a directory
above the data directory.) Null key values are not supported.

Multiple hashes can be tied to this module at the same time. 
The DEBUG and STACKSIZE values are global to this module, so
any changes will affect each of the tied hashes, not just the 
one calling the change.

Be aware of the fact that this module calls C<eval()> on the 
datafiles that it reads. You must take the necessary precautions 
to ensure the datafiles cannot be compromised, lest someone
leave you code that could damage your program. 

If your values contain perl references, you cannot modify them in place.
It appears that the C<tie()> function cannot handle it. For example,
given this code:

 tie %hash, 'Tie::PerlFiles', '/tmp/data', 'rw';
 $hash{'a'} = [ 1, 2, 3 ];

The following code will store bad data:

 $hash{'a'}->[1] = 4;

The correct way to handle this is to grab the value in another
pointer, modify it there, and then store it like so:

 $ref = $hash{'a'};
 $ref->[1] = 4;
 $hash{'a'} = $ref;

=head1 AUTHOR

Charles Brabec (brabec@ncsu.edu / charles@brabec.com)

=head1 VERSION

Version 1.3     25-Jan-2001

=head1 COPYRIGHT

Copyright (c) 2001 Charles Brabec. All rights reserved.

This library is free software; you can redistibute it and/or
modify it under the same terms as Perl itself.

=head1 SEE ALSO

perl(1), L<Data::Dumper>

=cut
